#g-policies-edit-form, #g-user-quota-form
  .g-quota-description
    font-size 12px
    margin 0 0 2px
  .g-quota-capacity-chart
    height 15px
    width 280px
    &.g-has-chart
      height 150px
  .g-sizeValue
    display inline-block
    width 150px
    margin-right 10px
  .g-sizeUnits
    display inline-block
    width auto
  .g-customQuota
    display inline-block
    margin 0 10px 0 0
